@@ -1,34 +1,30 @@
 #
 # Please see the .pod files for documentation. This module is copyrighted
 # as per the usual perl legalese:
-# Copyright (c) 1997 Austin Schutz.
-# expect() interface & functionality enhancements (c) 1999 Roland Giersig.
-#
-# All rights reserved. This program is free software; you can
-# redistribute it and/or modify it under the same terms as Perl
-# itself.
+# Copyright (c) 1997 Austin Schutz. All rights reserved. This program is free
+# software; you can redistribute it and/or modify it under the same terms as
+# Perl itself.
 #
 # Don't blame/flame me if you bust your stuff.
-# Austin Schutz <ASchutz@users.sourceforge.net>
+# Austin Schutz <tex@habit.com>
 #
-# This module now is maintained by
-# Roland Giersig <RGiersig@cpan.org>
+# expect() interface & functionality enhanced by
+# Roland Giersig <RGiersig@users.sourceforge.net>
 #
 
-use 5;				# 4 won't cut it. 
+require 5;			# 4 won't cut it. 
 
 package Expect;
 
-use IO::Pty 0.97;		# We need make_slave_controlling_terminal()
-use IO::Tty;
+use IO::Pty;			# This appears to require 5.004
+use IO::Stty;
 
 use strict 'refs';
 use strict 'vars';
 use strict 'subs';
 use POSIX;			# For setsid. 
 use Fcntl;			# For checking file handle settings.
-use Carp qw(cluck croak carp confess);
-use IO::Handle;
+use Carp;
 use Exporter;
 
 # This is necessary to make routines within Expect work.
@@ -37,7 +33,7 @@ use Exporter;
 @Expect::EXPORT = qw(expect exp_continue exp_continue_timeout);
 
 BEGIN {
-  $Expect::VERSION = "1.13_08";
+  $Expect::VERSION = "1.08";
   # These are defaults which may be changed per object, or set as
   # the user wishes.
   # This will be unset, since the default behavior differs between 
@@ -45,14 +41,11 @@ BEGIN {
   #  $Expect::Log_Stdout = 1;
   $Expect::Log_Group = 1;
   $Expect::Debug = 0;
-  $Expect::Exp_Max_Accum = 0; # unlimited
   $Expect::Exp_Internal = 0;
   $Expect::Manual_Stty = 0;
   $Expect::Multiline_Matching = 1;
-  $Expect::Do_Soft_Close = 0;
   @Expect::Before_List = ();
   @Expect::After_List = ();
-  %Expect::Spawned_PIDs = ();
 }
 
 sub version {
@@ -63,96 +56,52 @@ sub version {
   return $Expect::VERSION;
 }
 
-sub new {
+sub new { goto &spawn; }	# We should be called as Expect->spawn
+                                # or spawn Expect
 
+sub spawn {
+  my($tty);
+  my($name_of_tty);
   my ($class) = shift;
-  $class = ref($class) if ref($class); # so we can be called as $exp->new()
-
   # Create the pty which we will use to pass process info.
   my($self) = new IO::Pty;
-  die "$class: Could not assign a pty" unless $self;
-  bless $self => $class;
-  $self->autoflush(1);
+  bless ($self, $class);
+  my($cmd) = join(' ', @_);	# spawn is passed command line args.
+  ${*$self}{"exp_Command"} = $cmd;
+  $name_of_tty = $self->IO::Pty::ttyname();
+  die "$class: Could not assign a pty" unless $self->IO::Pty::ttyname();
+  $self->IO::Pty::autoflush();
 
   # This is defined here since the default is different for
   # initialized handles as opposed to spawned processes.
   ${*$self}{exp_Log_Stdout} = 1;
   $self->_init_vars();
-
-  if (@_) {
-    # we got add'l parms, so pass them to spawn
-    return $self->spawn(@_);
-  }
-  return $self;
-}
-
-sub spawn {
-  my ($class) = shift;
-  my $self;
-
-  if (ref($class)) {
-    $self = $class;
-  } else {
-    $self = $class->new();
-  }
-
-  my(@cmd) = @_;	# spawn is passed command line args.
-  ${*$self}{"exp_Command"} = \@cmd;
-
-  # set up pipe to detect childs exec error
-  pipe(STAT_RDR, STAT_WTR) or die "Cannot open pipe: $!";
-  STAT_WTR->autoflush(1);
-
-  my $pid = fork;
-
-  unless (defined ($pid)) {
-    warn "Cannot fork: $!" if $^W;
+  ${*$self}{exp_Pid} = fork;
+  unless (defined (${*$self}{exp_Pid})) {
+    warn "Cannot fork: $!";
     return undef;
   }
-
-  if($pid) {
-    # parent
-    my $errno;
-    ${*$self}{exp_Pid} = $pid;
-    close STAT_WTR;
-    $self->close_slave();
-    $self->set_raw() if $self->raw_pty and isatty($self);
-
-    # now wait for child exec (eof due to close-on-exit) or exec error
-    my $errstatus = sysread(STAT_RDR, $errno, 256);
-    die "Cannot sync with child: $!" if not defined $errstatus;
-    close STAT_RDR;
-    if ($errstatus) {
-      $! = $errno+0;
-      warn "Cannot exec(@cmd): $!\n" if $^W;
-      return undef;
-    }
-  }
-  else {
-    # child
-    close STAT_RDR;
-
-    $self->make_slave_controlling_terminal();
-    my $slv = $self->slave()
-      or die "Cannot get slave: $!";
-
-    $slv->set_raw() if $self->raw_pty;
+  unless (${*$self}{exp_Pid}) {
+    # Child
+    # Create a new 'session', lose controlling terminal.
+    POSIX::setsid() || warn "Couldn't perform setsid. Strange behavior may result.\r\n Problem: $!\r\n";
+    $tty = $self->IO::Pty::slave(); # Create slave handle.
+    $name_of_tty = $tty->ttyname();
+    # We have to close everything and then reopen ttyname after to get
+    # a controlling terminal.
     close($self);
-    close(STDIN);
-    open(STDIN,"<&". $slv->fileno())
-      or die "Couldn't reopen STDIN for reading, $!\n";
-    close(STDOUT);
-    open(STDOUT,">&". $slv->fileno())
-      or die "Couldn't reopen STDOUT for writing, $!\n";
-    close(STDERR);
-    open(STDERR,">&". $slv->fileno())
-      or die "Couldn't reopen STDERR for writing, $!\n";
+    close STDIN; close STDOUT;
+    open(STDIN,"<&". $tty->fileno()) || die "Couldn't reopen ". $name_of_tty ." for reading, $!\r\n";
+    open(STDOUT,">&". $tty->fileno()) || die "Couldn't reopen ". $name_of_tty ." for writing, $!\r\n";
+    close STDERR; # put that here or we would never see those die's above...
+    open(STDERR,">&". $tty->fileno()) || die "Couldn't redirect STDERR, $!\r\n";
 
-    { exec(@cmd) };
-    print STAT_WTR $!+0;
-    die "Cannot exec(@cmd): $!\n";
+    exec (@_);
+    # End Child.
   }
 
+  # Parent
+
   # This is sort of for code compatibility, and to make debugging a little
   # easier. By code compatibility I mean that previously the process's
   # handle was referenced by $process{Pty_Handle} instead of just $process.
@@ -160,13 +109,12 @@ sub spawn {
   # I think this also reflects Tcl Expect-like behavior.
   ${*$self}{exp_Pty_Handle} = "spawn id(".$self->fileno().")";
   if ((${*$self}{"exp_Debug"}) or (${*$self}{"exp_Exp_Internal"})) {
-    cluck("Spawned '@cmd'\r\n",
-	  "\t${*$self}{exp_Pty_Handle}\r\n",
-	  "\tPid: ${*$self}{exp_Pid}\r\n",
-	  "\tTty: ".$self->SUPER::ttyname()."\r\n",
-	 );
+    print STDERR ("Spawned '$cmd'\r\n",
+		  "\t${*$self}{exp_Pty_Handle}\r\n",
+		  "\tPid: ${*$self}{exp_Pid}\r\n",
+		  "\tTty: $name_of_tty\r\n",
+		 );
   }
-  $Expect::Spawned_PIDs{${*$self}{exp_Pid}} = undef;
   return $self;
 }
 
@@ -178,9 +126,9 @@ sub exp_init {
   my ($class) = shift;
   my($self) = shift;
   bless $self, $class;
-  croak "exp_init not passed a file object, stopped"
+  die "exp_init not passed a file object, stopped" 
     unless defined($self->fileno());
-  $self->autoflush(1);
+  $self->autoflush();
   # Define standard variables.. debug states, etc.
   $self->_init_vars();
   # Turn of logging. By default we don't want crap from a file to get spewed
@@ -193,47 +141,27 @@ sub exp_init {
   return $self;
 }
 
-# make an alias
-*init = \&exp_init;
 
 ######################################################################
 # We're happy OOP people. No direct access to stuff.
 # For standard read-writeable parameters, we define some autoload magic...
 my %Writeable_Vars = ( debug            => 'exp_Debug',
 		       exp_internal     => 'exp_Exp_Internal',
-		       do_soft_close    => 'exp_Do_Soft_Close',
-		       max_accum        => 'exp_Max_Accum',
-		       match_max        => 'exp_Max_Accum',
-		       notransfer       => 'exp_NoTransfer',
+		       exp_command      => 'exp_Command',
 		       log_stdout       => 'exp_Log_Stdout',
-		       log_user         => 'exp_Log_Stdout',
 		       log_group        => 'exp_Log_Group',
 		       manual_stty      => 'exp_Manual_Stty',
-		       restart_timeout_upon_receive => 'exp_Continue',
-		       raw_pty           => 'exp_Raw_Pty',
 		     );
 my %Readable_Vars = ( pid               => 'exp_Pid',
-		      exp_pid           => 'exp_Pid',
 		      exp_match_number  => 'exp_Match_Number',
-		      match_number      => 'exp_Match_Number',
 		      exp_error         => 'exp_Error',
-		      error             => 'exp_Error',
 		      exp_command       => 'exp_Command',
-		      command           => 'exp_Command',
 		      exp_match         => 'exp_Match',
-		      match             => 'exp_Match',
 		      exp_matchlist     => 'exp_Matchlist',
-		      matchlist         => 'exp_Matchlist',
 		      exp_before        => 'exp_Before',
-		      before            => 'exp_Before',
 		      exp_after         => 'exp_After',
-		      after             => 'exp_After',
 		      exp_exitstatus    => 'exp_Exit',
-		      exitstatus        => 'exp_Exit',
 		      exp_pty_handle    => 'exp_Pty_Handle',
-		      pty_handle        => 'exp_Pty_Handle',
-		      exp_logfile       => 'exp_Log_File',
-		      logfile           => 'exp_Log_File',
 		      %Writeable_Vars,
 		    );
 
@@ -250,12 +178,11 @@ sub AUTOLOAD {
     croak "ERROR: cannot find method `$name' in class $type";
   }
   my $varname = $Readable_Vars{$name};
-  my $tmp;
-  $tmp = ${*$self}{$varname} if exists ${*$self}{$varname};
+  my $tmp = ${*$self}{$varname};
 
   if (@_) {
     if (exists $Writeable_Vars{$name}) {
-      my $ref = ref($tmp);
+      my $ref = ref(${*$self}{$varname});
       if ($ref eq 'ARRAY') {
 	${*$self}{$varname} = [ @_ ];
       } elsif ($ref eq 'HASH') {
@@ -313,7 +240,7 @@ sub set_group {
     warn("Attempting to set a handle group on ${*$self}{exp_Pty_Handle}, ",
 	 "a non-readable handle!\r\n");
   }
-  while ($write_handle = shift) {
+  while ($write_handle = shift) { 
     if ($write_handle->_get_mode() !~ 'w') {
       warn("Attempting to set a non-writeable listen handle ",
 	   "${*$write_handle}{exp_Pty_handle} for ",
@@ -323,84 +250,47 @@ sub set_group {
   }
 }
 
-sub log_file {
-    my $self = shift;
-
-    return(${*$self}{exp_Log_File})
-      if not @_;  # we got no param, return filehandle
-
-    my $file = shift;
-    my $mode = shift || "a";
-
-    if (${*$self}{exp_Log_File} and ref(${*$self}{exp_Log_File}) ne 'CODE') {
-      close(${*$self}{exp_Log_File});
-      ${*$self}{exp_Log_File} = undef;
-    }
-    return if (not $file);
-    my $fh = $file;
-    if (not ref($file)) {
-      # it's a filename
-      $fh = new IO::File $file, $mode
-	or croak "Cannot open logfile $file: $!";
-    }
-    if (ref($file) ne 'CODE') {
-      croak "Given logfile doesn't have a 'print' method"
-	if not $fh->can("print");
-      $fh->autoflush(1);		# so logfile is up to date
-    }
-
-    ${*$self}{exp_Log_File} = $fh;
+sub max_accum {
+  my($self) = shift;
+  return ${*$self}{"exp_Max_Accum"} unless defined($_[0]);
+  ${*$self}{"exp_Max_Accum"} = shift;
+  if (${*$self}{"exp_Max_Accum"} eq 'undefine') {
+    undef(${*$self}{"exp_Max_Accum"});
+  }
 }
 
-
 # I'm going to leave this here in case I might need to change something.
 # Previously this was calling `stty`, in a most bastardized manner.
 sub exp_stty {
   my($self) = shift;
-  my($mode) = "@_";
-  
+  my($mode) = shift;
   return undef unless defined($mode);
-  if (not defined $INC{"IO/Stty.pm"}) {
-    carp "IO::Stty not installed, cannot change mode";
-    return undef;
-  }
-
   if (${*$self}{"exp_Debug"}) {
     print STDERR "Setting ${*$self}{exp_Pty_Handle} to tty mode '$mode'\r\n";
   }
   unless (POSIX::isatty($self)) {
-    if (${*$self}{"exp_Debug"} or $^W) {
-      warn "${*$self}{exp_Pty_Handle} is not a tty. Not changing mode";
+    if (${*$self}{"exp_Debug"}) {
+      print STDERR "${*$self}{exp_Pty_Handle} is not a tty. Not changing mode\r\n";
     }
     return '';			# No undef to avoid warnings elsewhere.
   }
-  IO::Stty::stty($self, split(/\s/,$mode));
+  IO::Stty::stty($self,split(/\s/,$mode));
 }
 
-*stty = \&exp_stty;
-
 # If we want to clear the buffer. Otherwise Accum will grow during send_slow
 # etc. and contain the remainder after matches.
 sub clear_accum {
-  my ($self) = shift;
+  my($self) = shift;
   my ($temp) = (${*$self}{exp_Accum});
   ${*$self}{exp_Accum} = '';
   # return the contents of the accumulator.
   return $temp;
 }
 
-sub set_accum {
-  my ($self) = shift;
-  my ($temp) = (${*$self}{exp_Accum});
-  ${*$self}{exp_Accum} = shift;
-  # return the contents of the accumulator.
-  return $temp;
-}
-
 ######################################################################
 # define constants for pattern subs
-sub exp_continue() { "exp_continue" }
-sub exp_continue_timeout() { "exp_continue_timeout" }
+sub exp_continue() { 0xDEADBEEF }
+sub exp_continue_timeout() { 0xBADDEAD }
 
 ######################################################################
 # Expect on multiple objects at once.
@@ -423,8 +313,7 @@ sub exp_continue_timeout() { "exp_continue_timeout" }
 
 sub expect {
   my $self;
-  $self = shift if (ref($_[0]) and $_[0]->isa('Expect'));
-  shift if defined $_[0] and $_[0] eq 'Expect';	# or as Expect->expect
+  $self = shift if (ref($_[0])); # can also be called as Expect::expect
   my $timeout = shift;
   my $timeout_hook = undef;
 
@@ -527,7 +416,7 @@ sub expect {
     _multi_expect(0, undef, @Expect::Before_List);
   }
 
-  cluck ("Starting EXPECT pattern matching...\r\n")
+  print STDERR ("Starting EXPECT pattern matching...\r\n")
     if ($debug or $internal);
   my @ret;
   @ret = _multi_expect($timeout, $timeout_hook,
@@ -558,7 +447,7 @@ sub _multi_expect($$@) {
   foreach my $pat (@_) {
     my @patterns = @{$pat}[1..$#{$pat}];
     foreach my $exp (@{$pat->[0]}) {
-      if (exists ${*$exp}{"exp_Max_Accum"} and ${*$exp}{"exp_Max_Accum"}) {
+      if (defined(${*$exp}{"exp_Max_Accum"})) {
 	${*$exp}{exp_Accum} =
 	  $exp->_trim_length(${*$exp}{exp_Accum},
 			     ${*$exp}{exp_Max_Accum});
@@ -567,11 +456,11 @@ sub _multi_expect($$@) {
 		    "\tTimeout: ",
 		    (defined($timeout) ? $timeout : "unlimited" ),
 		    " seconds.\r\n",
-		    "\tCurrent time: ". localtime(). "\r\n",
+		    "\tCurrent time: ", time(), "\r\n",
 		   ) if $Expect::Debug;
 
       # What are we expecting? What do you expect? :-)
-      if (${*$exp}{exp_Exp_Internal}) {
+      if ($Expect::Exp_Internal) {
 	print STDERR "${*$exp}{exp_Pty_Handle}: list of patterns:\r\n";
 	foreach my $pattern (@patterns) {
 	  print STDERR ('  ',
@@ -597,13 +486,11 @@ sub _multi_expect($$@) {
 
   # Set the last loop time to now for time comparisons at end of loop.
   my $start_loop_time = time();
-  my $exp_cont = 1;
 
  READLOOP:
-  while ($exp_cont) {
-    $exp_cont = 1;
-    $err = "";
+  while (1) {
     my $rmask = '';
+    my $exp_cont = '';
     my $time_left = undef;
     if (defined $timeout) {
       $time_left = $timeout - (time() - $start_loop_time);
@@ -634,7 +521,7 @@ sub _multi_expect($$@) {
 	print STDERR ("\r\n${*$exp}{exp_Pty_Handle}: Does `",
 		      $exp->_trim_length(_make_readable(${*$exp}{exp_Accum})),
 		      "'\r\nmatch:\r\n")
-	  if ${*$exp}{exp_Exp_Internal};
+	  if $Expect::Exp_Internal;
 
 	# we don't keep the parameter number anymore
 	# (clashes with before & after), instead the parameter number is
@@ -646,7 +533,7 @@ sub _multi_expect($$@) {
 			": ", $pattern->[1],
 			" `", _make_readable($pattern->[2]),
 			"'? ")
-	    if (${*$exp}{exp_Exp_Internal});
+	    if ($Expect::Exp_Internal);
 
 	  # Matching exactly
 	  if ($pattern->[1] eq '-ex') {
@@ -696,10 +583,9 @@ sub _multi_expect($$@) {
 	  }
 
 	  if ($exp_matched) {
-	    ${*$exp}{exp_Accum} = $after
-	      unless ${*$exp}{exp_NoTransfer};
+	    ${*$exp}{exp_Accum} = $after;
 	    print STDERR "YES!!\r\n"
-	      if ${*$exp}{exp_Exp_Internal};
+	      if $Expect::Exp_Internal;
 	    print STDERR ("    Before match string: `",
 			  $exp->_trim_length(_make_readable(($before))),
 			  "'\r\n",
@@ -716,12 +602,12 @@ sub _multi_expect($$@) {
 				   @matchlist),
 			       ),
 			  ")\r\n",
-			 ) if (${*$exp}{exp_Exp_Internal});
+			 ) if ($Expect::Exp_Internal);
 
 	    # call hook function if defined
 	    if ($pattern->[3]) {
 	      print STDERR ("Calling hook $pattern->[3]...\r\n",
-			   ) if (${*$exp}{exp_Exp_Internal} or $Expect::Debug);
+			   ) if ($Expect::Exp_Internal or $Expect::Debug);
 	      if ($#{$pattern} > 3) {
 		# call with parameters if given
 		$exp_cont = &{$pattern->[3]}($exp,
@@ -730,21 +616,26 @@ sub _multi_expect($$@) {
 		$exp_cont = &{$pattern->[3]}($exp);
 	      }
 	    }
-	    if ($exp_cont and $exp_cont eq exp_continue) {
-	      print STDERR ("Continuing expect, restarting timeout...\r\n")
-		if (${*$exp}{exp_Exp_Internal} or $Expect::Debug);
-	      $start_loop_time = time(); # restart timeout count
-	      next READLOOP;
-	    } elsif ($exp_cont and $exp_cont eq exp_continue_timeout) {
-	      print STDERR ("Continuing expect...\r\n")
-		if (${*$exp}{exp_Exp_Internal} or $Expect::Debug);
+	    if ($exp_cont == exp_continue
+		or $exp_cont eq 'exp_continue'
+		or $exp_cont eq 'exp_continue_timeout'
+		or $exp_cont == exp_continue_timeout) {
+	      if ($exp_cont == exp_continue
+		  or $exp_cont eq 'exp_continue') {
+		print STDERR ("Continuing expect, restarting timeout...\r\n",
+			     ) if ($Expect::Exp_Internal or $Expect::Debug);
+		$start_loop_time = time() # restart timeout count
+	      } else {
+		print STDERR ("Continuing expect...\r\n",
+			     ) if ($Expect::Exp_Internal or $Expect::Debug);
+	      }
 	      next READLOOP;
 	    }
 	    last READLOOP;
 	  }
-	  print STDERR "No.\r\n" if ${*$exp}{exp_Exp_Internal};
+	  print STDERR "No.\r\n" if $Expect::Exp_Internal;
 	}
-	print STDERR "\r\n" if ${*$exp}{exp_Exp_Internal};
+	print STDERR "\r\n" if $Expect::Exp_Internal;
 	# don't have to match again until we get new data
 	${*$exp}{exp_New_Data} = 0;
       }
@@ -758,39 +649,7 @@ sub _multi_expect($$@) {
     my $nfound = select($rmask, undef, undef, $time_left);
 
     # go until we don't find something (== timeout).
-    if (not $nfound) {
-      # No pattern, no EOF. Did we time out?
-      $err = "1:TIMEOUT";
-      foreach my $pat (@_) {
-	foreach my $exp (@{$pat->[0]}) {
-	  $before = ${*$exp}{exp_Before} = ${*$exp}{exp_Accum};
-	  next if not defined $exp->fileno(); # skip already closed
-	  ${*$exp}{exp_Error} = $err unless ${*$exp}{exp_Error};
-	}
-      }
-      print STDERR ("TIMEOUT\r\n")
-	if ($Expect::Debug or $Expect::Exp_Internal);
-      if ($timeout_hook) {
-	my $ret;
-	print STDERR ("Calling timeout function $timeout_hook...\r\n")
-	  if ($Expect::Debug or $Expect::Exp_Internal);
-	if (ref($timeout_hook) eq 'CODE') {
-	  $ret = &{$timeout_hook}($_[0]->[0]);
-	} else {
-	  if ($#{$timeout_hook} > 3) {
-	    $ret = &{$timeout_hook->[3]}($_[0]->[0],
-					 @{$timeout_hook}[4..$#{$timeout_hook}]);
-	  } else {
-	    $ret = &{$timeout_hook->[3]}($_[0]->[0]);
-	  }
-	}
-	if ($ret and $ret eq exp_continue) {
-	  $start_loop_time = time();	# restart timeout count
-	  next READLOOP;
-	}
-      }
-      last READLOOP;
-    }
+    last READLOOP unless ($nfound);
 
     my @bits = split(//,unpack('b*',$rmask));
     foreach my $pat (@_) {
@@ -809,15 +668,11 @@ sub _multi_expect($$@) {
 	  if ($nread == 0) {
 	    print STDERR ("${*$exp}{exp_Pty_Handle}: EOF\r\n")
 	      if ($Expect::Debug);
-	    $before = ${*$exp}{exp_Before} = $exp->clear_accum();
+	    ${*$exp}{exp_Before} = $exp->clear_accum();
 	    $err = "2:EOF";
 	    ${*$exp}{exp_Error} = $err;
-	    ${*$exp}{exp_Has_EOF} = 1;
-	    $exp_cont = undef;
 	    foreach my $eof_pat (grep {$_->[1] eq '-eof'} @{$pat}[1..$#{$pat}]) {
 	      my $ret;
-	      print STDERR ("Calling EOF hook $eof_pat->[3]...\r\n",
-			   ) if ($Expect::Debug);
 	      if ($#{$eof_pat} > 3) {
 		# call with parameters if given
 		$ret = &{$eof_pat->[3]}($exp,
@@ -825,10 +680,11 @@ sub _multi_expect($$@) {
 	      } else {
 		$ret = &{$eof_pat->[3]}($exp);
 	      }
-	      if ($ret and
-		  ($ret eq exp_continue
-		   or $ret eq exp_continue_timeout)) {
-		    $exp_cont = $ret;
+	      if ($ret == exp_continue
+		  or $ret eq 'exp_continue'
+		  or $ret eq 'exp_continue_timeout'
+		  or $ret == exp_continue_timeout) {
+		$exp_cont = $ret;
 	      }
 	    }
 	    # is it dead?
@@ -840,7 +696,6 @@ sub _multi_expect($$@) {
 		$err = "3:Child PID ${*$exp}{exp_Pid} exited with status $?";
 		${*$exp}{exp_Error} = $err;
 		${*$exp}{exp_Exit} = $?;
-		delete $Expect::Spawned_PIDs{${*$exp}{exp_Pid}};
 		${*$exp}{exp_Pid} = undef;
 	      }
 	    }
@@ -857,29 +712,49 @@ sub _multi_expect($$@) {
 	  $buffer =~ s/ \cH//g;
 	  # Append it to the accumulator.
 	  ${*$exp}{exp_Accum} .= $buffer;
-	  if (exists ${*$exp}{exp_Max_Accum}
-	      and ${*$exp}{exp_Max_Accum}) {
-	    ${*$exp}{exp_Accum} =
-	      $exp->_trim_length(${*$exp}{exp_Accum},
-				 ${*$exp}{exp_Max_Accum});
-	  }
 	  ${*$exp}{exp_New_Data} = 1; # next round we try to match again
 
-	  $exp_cont = exp_continue
-	    if (exists ${*$exp}{exp_Continue} and ${*$exp}{exp_Continue});
 	  # Now propagate what we have read to other listeners...
 	  $exp->_print_handles($buffer);
 
 	  # End handle reading section.
 	}
+
       }
     }				# end read loop
     $start_loop_time = time()	# restart timeout count
-      if ($exp_cont and $exp_cont eq exp_continue);
+      if ($exp_cont eq exp_continue or $exp_cont eq 'exp_continue');
   }
   # End READLOOP
 
   # Post loop. Do we have anything?
+  # No pattern, no EOF. Did we time out or is the process dead?
+  if (not $exp_matched) {
+    $err = "1:TIMEOUT";
+    foreach my $pat (@_) {
+      foreach my $exp (@{$pat->[0]}) {
+	next if not defined $exp->fileno(); # skip already closed
+	${*$exp}{exp_Error} = $err unless ${*$exp}{exp_Error};
+      }
+    }
+    print STDERR ("TIMEOUT\r\n")
+      if ($Expect::Debug or $Expect::Exp_Internal);
+    if ($timeout_hook) {
+      print STDERR ("Calling timeout function $timeout_hook...\r\n")
+	if ($Expect::Debug or $Expect::Exp_Internal);
+      if (ref($timeout_hook) eq 'CODE') {
+	&{$timeout_hook}($_[0]->[0]);
+      } else {
+	if ($#{$timeout_hook} > 3) {
+	  &{$timeout_hook->[3]}($_[0]->[0],
+				@{$timeout_hook}[4..$#{$timeout_hook}]);
+	} else {
+	  &{$timeout_hook->[3]}($_[0]->[0]);
+	}
+      }
+    }
+  }
+
   # Tell us status
   if ($Expect::Debug or $Expect::Exp_Internal) {
       if ($exp_matched) {
@@ -890,7 +765,7 @@ sub _multi_expect($$@) {
 			"\r\n  Error: ${*$exp_matched}{exp_Error}." : '',
 			"\r\n");
       } else {
-	  print STDERR ("Returning from expect with TIMEOUT or EOF\r\n");
+	  print STDERR ("Returning from expect with TIMEOUT\r\n");
       }
     if ($Expect::Debug and $exp_matched) {
       print STDERR "  ${*$exp_matched}{exp_Pty_Handle}: accumulator: `";
@@ -918,7 +793,7 @@ sub _multi_expect($$@) {
 	${*$exp_matched}{exp_Match_Number};
   }
 
-  return wantarray? (undef, $err, undef, $before, undef, undef) : undef;
+  return wantarray? (undef, undef, undef, undef, undef, undef) : undef;
 }
 
 
@@ -1035,7 +910,7 @@ sub interact {
   # Normally an undef passed to set_group will return the current groups.
   # It is possible that it may be of worth to make it possible to undef
   # The current group without doing this.
-  unless (@old_group) {
+  unless (defined (@old_group)) {
     @{${*$self}{exp_Listen_Group}} = ();
   }
   $self->manual_stty($old_manual_stty_val);
@@ -1047,7 +922,7 @@ sub interconnect {
   #  my ($handle)=(shift); call as Expect::interconnect($spawn1,$spawn2,...)
   my ($rmask,$nfound,$nread);
   my ($rout, @bits, $emask, $eout, @ebits ) = ();
-  my ($escape_sequence,$escape_character_buffer);
+  my ($escape_sequence,$escape_character_buffer,$offset);
   my (@handles) = @_;
   my ($handle,$read_handle,$write_handle);
   my ($read_mask,$temp_mask) = ('','');
@@ -1103,13 +978,9 @@ sub interconnect {
   while (1) {
     # test each handle to see if it's still alive.
     foreach $read_handle (@handles) {
-      waitpid(${*$read_handle}{exp_Pid}, WNOHANG)
-	if (exists (${*$read_handle}{exp_Pid}) and ${*$read_handle}{exp_Pid});
-      if (exists(${*$read_handle}{exp_Pid})
-	  and (${*$read_handle}{exp_Pid})
-	  and (! kill(0,${*$read_handle}{exp_Pid}))) {
-	print STDERR "Got EOF (${*$read_handle}{exp_Pty_Handle} died) reading ${*$read_handle}{exp_Pty_Handle}\r\n"
-	  if ${*$read_handle}{"exp_Debug"};
+      waitpid(${*$read_handle}{exp_Pid}, WNOHANG) if (defined (${*$read_handle}{exp_Pid}) &&${*$read_handle}{exp_Pid});
+      if (defined(${*$read_handle}{exp_Pid}) &&(${*$read_handle}{exp_Pid}) &&(! kill(0,${*$read_handle}{exp_Pid}))) {
+	print STDERR "Got EOF (${*$read_handle}{exp_Pty_Handle} died) reading ${*$read_handle}{exp_Pty_Handle}\r\n"if ${*$read_handle}{"exp_Debug"};
 	last CONNECT_LOOP unless defined(${${*$read_handle}{exp_Function}}{"EOF"});
 	last CONNECT_LOOP unless &{${${*$read_handle}{exp_Function}}{"EOF"}}(@{${${*$read_handle}{exp_Parameters}}{"EOF"}});
       }
@@ -1137,7 +1008,10 @@ sub interconnect {
 	foreach $escape_sequence (keys(%{${*$read_handle}{exp_Function}})) {
 	  print STDERR "Tested escape sequence $escape_sequence from ${*$read_handle}{exp_Pty_Handle}"if ${*$read_handle}{"exp_Debug"} > 1;
 	  # Make sure it doesn't grow out of bounds.
-	  $escape_character_buffer = $read_handle->_trim_length($escape_character_buffer,${*$read_handle}{"exp_Max_Accum"}) if (${*$read_handle}{"exp_Max_Accum"});
+	  if ((defined(${*$read_handle}{"exp_Max_Accum"}) &&(length($escape_character_buffer)) > ${*$read_handle}{"exp_Max_Accum"})) {
+	    $offset = length($escape_character_buffer) - ${*$read_handle}{"exp_Max_Accum"};
+	    $escape_character_buffer = substr($escape_character_buffer,$offset,${*$read_handle}{"exp_Max_Accum"});
+	  }
 	  if ($escape_character_buffer =~ /($escape_sequence)/) {
 	    if (${*$read_handle}{"exp_Debug"}) {
 	      print STDERR "\r\ninterconnect got escape sequence from ${*$read_handle}{exp_Pty_Handle}.\r\n";
@@ -1197,37 +1071,6 @@ sub interconnect {
   }
 }
 
-# user can decide if log output gets also sent to logfile
-sub print_log_file {
-  my $self = shift;
-  if (${*$self}{exp_Log_File}) {
-    if (ref(${*$self}{exp_Log_File}) eq 'CODE') {
-      ${*$self}{exp_Log_File}->(@_);
-    } else {
-      ${*$self}{exp_Log_File}->print(@_);
-    }
-  }
-}
-
-# we provide our own print so we can debug what gets sent to the
-# processes...
-sub print (@) {
-  my ($self, @args) = @_;
-  if (${*$self}{exp_Exp_Internal}) {
-    my $args = _make_readable(join('', @args));
-    cluck "Sending '$args' to ${*$self}{exp_Pty_Handle}\r\n";
-  }
-  foreach my $arg (@args) {
-    while (length($arg) > 80) {
-      $self->SUPER::print(substr($arg, 0, 80));
-      $arg = substr($arg, 80);
-    }
-    $self->SUPER::print($arg);
-  }
-}
-
-# make an alias for Tcl/Expect users for a DWIM experience...
-*send = \&print;
 
 # This is an Expect standard. It's nice for talking to modems and the like
 # where from time to time they get unhappy if you send items too quickly.
@@ -1255,7 +1098,7 @@ sub send_slow{
 	  # Is this necessary to keep? Probably.. #
 	  # if you need to expect it later.
 	  ${*$self}{exp_Accum}.= ${*$self}{exp_Pty_Buffer};
-	  ${*$self}{exp_Accum} = $self->_trim_length(${*$self}{exp_Accum},${*$self}{"exp_Max_Accum"}) if (${*$self}{"exp_Max_Accum"});
+	  ${*$self}{exp_Accum} = $self->_trim_length(${*$self}{exp_Accum},${*$self}{"exp_Max_Accum"}) if defined (${*$self}{"exp_Max_Accum"});
 	  $self->_print_handles(${*$self}{exp_Pty_Buffer});
 	  print STDERR "Received \'".$self->_trim_length(_make_readable($char))."\' from ${*$self}{exp_Pty_Handle}\r\n" if ${*$self}{"exp_Debug"} > 1;
 	}
@@ -1302,48 +1145,43 @@ sub soft_close {
   my($end_time,$select_time,$temp_buffer);
   my($close_status);
   # Give it 15 seconds to cough up an eof.
-  cluck "Closing ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
+  print STDERR "Closing ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
   return -1 if not defined $self->fileno(); # skip if handle already closed
-  unless (exists ${*$self}{exp_Has_EOF} and ${*$self}{exp_Has_EOF}) {
-    $end_time = time() + 15;
-    while ($end_time > time()) {
-      $select_time = $end_time - time();
-      # Sanity check.
-      $select_time = 0 if $select_time < 0;
-      $rmask = '';
-      vec($rmask,$self->fileno(),1) = 1;
-      ($nfound) = select($rmask,undef,undef,$select_time);
-      last unless (defined($nfound) && $nfound);
-      $nread = sysread($self,$temp_buffer,8096);
-      # 0 = EOF.
-      unless (defined($nread) && $nread) {
-	print STDERR "Got EOF from ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
-	last;
-      }
-      $self->_print_handles($temp_buffer);
-    }
-    if (($end_time <= time()) && ${*$self}{exp_Debug}) {
-      print STDERR "Timed out waiting for an EOF from ${*$self}{exp_Pty_Handle}.\r\n";
+  $end_time = time() + 15;
+  while ($end_time > time()) {
+    $select_time = $end_time - time();
+    # Sanity check.
+    $select_time = 0 if $select_time < 0;
+    $rmask = '';
+    vec($rmask,$self->fileno(),1) = 1;
+    ($nfound) = select($rmask,undef,undef,$select_time);
+    last unless (defined($nfound) && $nfound);
+    $nread = sysread($self,$temp_buffer,8096);
+    # 0 = EOF.
+    unless (defined($nread) && $nread) {
+      print STDERR "Got EOF from ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
+      last;
     }
+    $self->_print_handles($temp_buffer);
+  }
+  if (($end_time <= time()) && ${*$self}{exp_Debug}) {
+    print STDERR "Timed out waiting for an EOF from ${*$self}{exp_Pty_Handle}.\r\n";
   }
   if ( ($close_status = $self->close()) && ${*$self}{exp_Debug}) {
     print STDERR "${*$self}{exp_Pty_Handle} closed.\r\n";
   }
   # quit now if it isn't a process.
-  return $close_status unless defined(${*$self}{exp_Pid});
+  return $close_status unless defined($self->pid());
   # Now give it 15 seconds to die.
   $end_time = time() + 15;
   while ($end_time > time()) {
-    $returned_pid = waitpid(${*$self}{exp_Pid}, &WNOHANG);
+    $returned_pid = waitpid($self->pid(), &WNOHANG);
     # Stop here if the process dies.
     if (defined($returned_pid) && $returned_pid) {
-      delete $Expect::Spawned_PIDs{$returned_pid};
       if (${*$self}{exp_Debug}) {
 	print STDERR "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} exited, Status: $?\r\n";
       }
-      ${*$self}{exp_Pid} = undef;
-      ${*$self}{exp_Exit} = $?;
-      return ${*$self}{exp_Exit};
+      return $?;
     }
     sleep 1;			# Keep loop nice.
   }
@@ -1351,18 +1189,15 @@ sub soft_close {
   if (${*$self}{exp_Debug}) {
     print STDERR "${*$self}{exp_Pty_Handle} not exiting, sending TERM.\r\n";
   }
-  kill TERM => ${*$self}{exp_Pid};
+  kill(15,${*$self}{exp_Pid});
   # Now to be anal retentive.. wait 15 more seconds for it to die.
   $end_time = time() + 15;
   while ($end_time > time()) {
-    $returned_pid = waitpid(${*$self}{exp_Pid}, &WNOHANG);
+    $returned_pid = waitpid($self->pid(), &WNOHANG);
     if (defined($returned_pid) && $returned_pid) {
-      delete $Expect::Spawned_PIDs{$returned_pid};
       if (${*$self}{exp_Debug}) {
 	print STDERR "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} terminated, Status: $?\r\n";
       }
-      ${*$self}{exp_Pid} = undef;
-      ${*$self}{exp_Exit} = $?;
       return $?;
     }
     sleep 1;
@@ -1377,26 +1212,23 @@ sub hard_close {
   my($nfound,$nread,$rmask,$returned_pid);
   my($end_time,$select_time,$temp_buffer);
   my($close_status);
-  cluck "Closing ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
+  print STDERR "Closing ${*$self}{exp_Pty_Handle}.\r\n" if ${*$self}{exp_Debug};
   # Don't wait for an EOF.
   if ( ($close_status = $self->close()) && ${*$self}{exp_Debug}) {
     print STDERR "${*$self}{exp_Pty_Handle} closed.\r\n";
   }
   # Return now if handle.
-  return $close_status unless defined(${*$self}{exp_Pid});
+  return $close_status unless defined($self->pid());
   # Now give it 5 seconds to die. Less patience here if it won't die.
   $end_time = time() + 5;
   while ($end_time > time()) {
-    $returned_pid = waitpid(${*$self}{exp_Pid}, &WNOHANG);
+    $returned_pid = waitpid($self->pid(), &WNOHANG);
     # Stop here if the process dies.
     if (defined($returned_pid) && $returned_pid) {
-      delete $Expect::Spawned_PIDs{$returned_pid};
       if (${*$self}{exp_Debug}) {
 	print STDERR "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} exited, Status: $?\r\n";
       }
-      ${*$self}{exp_Pid} = undef;
-      ${*$self}{exp_Exit} = $?;
-      return ${*$self}{exp_Exit};
+      return $?;
     }
     sleep 1;			# Keep loop nice.
   }
@@ -1404,40 +1236,33 @@ sub hard_close {
   if (${*$self}{exp_Debug}) {
     print STDERR "${*$self}{exp_Pty_Handle} not exiting, sending TERM.\r\n";
   }
-  kill TERM => ${*$self}{exp_Pid};
+  kill(15,${*$self}{exp_Pid});
   # wait 15 more seconds for it to die.
   $end_time = time() + 15;
   while ($end_time > time()) {
-    $returned_pid = waitpid(${*$self}{exp_Pid}, &WNOHANG);
+    $returned_pid = waitpid($self->pid(), &WNOHANG);
     if (defined($returned_pid) && $returned_pid) {
-      delete $Expect::Spawned_PIDs{$returned_pid};
       if (${*$self}{exp_Debug}) {
 	print STDERR "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} terminated, Status: $?\r\n";
       }
-      ${*$self}{exp_Pid} = undef;
-      ${*$self}{exp_Exit} = $?;
-      return ${*$self}{exp_Exit};
+      return $?;
     }
     sleep 1;
   }
-  kill KILL => ${*$self}{exp_Pid};
+  kill(9,${*$self}{exp_Pid});
   # wait 5 more seconds for it to die.
   $end_time = time() + 5;
   while ($end_time > time()) {
-    $returned_pid = waitpid(${*$self}{exp_Pid}, &WNOHANG);
+    $returned_pid = waitpid($self->pid(), &WNOHANG);
     if (defined($returned_pid) && $returned_pid) {
-      delete $Expect::Spawned_PIDs{$returned_pid};
       if (${*$self}{exp_Debug}) {
 	print STDERR "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} killed, Status: $?\r\n";
       }
-      ${*$self}{exp_Pid} = undef;
-      ${*$self}{exp_Exit} = $?;
-      return ${*$self}{exp_Exit};
+      return $?;
     }
     sleep 1;
   }
   warn "Pid ${*$self}{exp_Pid} of ${*$self}{exp_Pty_Handle} is HUNG.\r\n";
-  ${*$self}{exp_Pid} = undef;
   return undef;
 }
 
@@ -1447,22 +1272,19 @@ sub _init_vars {
   my($self) = shift;
 
   # for every spawned process or filehandle.
-  ${*$self}{exp_Log_Stdout} = $Expect::Log_Stdout
+  ${*$self}{"exp_Log_Stdout"} = $Expect::Log_Stdout
     if defined ($Expect::Log_Stdout);
-  ${*$self}{exp_Log_Group} = $Expect::Log_Group;
-  ${*$self}{exp_Debug} = $Expect::Debug;
-  ${*$self}{exp_Exp_Internal} = $Expect::Exp_Internal;
-  ${*$self}{exp_Manual_Stty} = $Expect::Manual_Stty;
+  ${*$self}{"exp_Log_Group"} = $Expect::Log_Group;
+  ${*$self}{"exp_Debug"} = $Expect::Debug;
+  ${*$self}{"exp_Exp_Internal"} = $Expect::Exp_Internal;
+  ${*$self}{"exp_Manual_Stty"} = $Expect::Manual_Stty;
   ${*$self}{exp_Stored_Stty} = 'sane';
-  ${*$self}{exp_Do_Soft_Close} = $Expect::Do_Soft_Close;
 
   # sysread doesn't like my or local vars.
   ${*$self}{exp_Pty_Buffer} = '';
 
   # Initialize accumulator.
-  ${*$self}{exp_Max_Accum} = $Expect::Exp_Max_Accum;
   ${*$self}{exp_Accum} = '';
-  ${*$self}{exp_NoTransfer} = 0;
 
   # create empty expect_before & after lists
   ${*$self}{exp_expect_before_list} = [];
@@ -1499,10 +1321,10 @@ sub _trim_length {
   my($string) = shift;
   my($length) = shift;
 
-  # If we're not passed a length (_trim_length is being used for debugging
+  # If we're not passed a length (_trim_length is being used for debugging 
   # purposes) AND debug >= 3, don't trim.
-  return($string) if (defined ($self) and
-		      ${*$self}{"exp_Debug"} >= 3 and (!(defined($length))));
+  return($string) if (defined ($self) && 
+		      ${*$self}{"exp_Debug"} >= 3 && (!(defined($length))));
   my($indicate_truncation) = '...' unless $length;
   $length = 1021 unless $length;
   return($string) unless $length < length($string);
@@ -1527,10 +1349,8 @@ sub _print_handles {
     }
   }
   # If ${*$self}{exp_Pty_Handle} is STDIN this would make it echo.
-  print STDOUT $print_this
-    if ${*$self}{"exp_Log_Stdout"};
-  $self->print_log_file($print_this);
-  $|= 1; # This should not be necessary but autoflush() doesn't always work.
+  print $print_this if ${*$self}{"exp_Log_Stdout"};
+  $|= 1;			# This should not be necessary but autoflush() doesn't always work.
 }
 
 sub _get_mode {
@@ -1560,10 +1380,8 @@ sub _undef {
 # clean up child processes
 sub DESTROY {
   my $self = shift;
-  if (${*$self}{exp_Do_Soft_Close}) {
-    $self->soft_close();
-  }
-  $self->hard_close();
+  $self->hard_close()
+    unless defined $self->soft_close();
 }
 
 1;
